# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2019 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Daniel Izquierdo Cortazar <dizquierdo@bitergia.com>
#

from datetime import datetime as dt

import pandas

from grimoirelab_toolkit.datetime import str_to_datetime


class Events(object):
    """ Class that 'eventizes' information for a given dataset.

    This class aims at providing with some granularity all of the events
    of a pre-formatted item. This is expected to help in later steps
    during the visualization platform.
    """

    META_TIMESTAMP = "metadata__timestamp"
    META_UPDATED_ON = "metadata__updated_on"
    META_ENRICHED_ON = "metadata__enriched_on"

    GRIMOIRE_CREATION_DATE = "grimoire_creation_date"
    PROJECT = "project"
    PROJECT_1 = "project_1"
    PERCEVAL_UUID = "perceval_uuid"

    SH_AUTHOR_ID = "author_id"
    SH_AUTHOR_ORG_NAME = "author_org_name"
    SH_AUTHOR_NAME = "author_name"
    SH_AUTHOR_UUID = "author_uuid"
    SH_AUTHOR_DOMAIN = "author_domain"
    SH_AUTHOR_USER_NAME = "author_user_name"
    SH_AUTHOR_BOT = "author_bot"
    SH_AUTHOR_MULTI_ORG_NAMES = "author_multi_org_names"

    UNKNOWN = 'Unknown'

    def __init__(self, items, enrich):
        """ Main constructor of the class

        :param items: original list of JSON that contains all info about a commit
        :type items: list
        :param enrich:
        :type enrich: grimoire_elk.elk.enrich.Enrich
        """

        self.items = items
        self.enrich = enrich

    def _add_metadata(self, df_columns, item):
        metadata__timestamp = item["metadata__timestamp"]
        metadata__updated_on = item["metadata__updated_on"]
        metadata__enriched_on = dt.utcnow().isoformat()

        df_columns[Events.META_TIMESTAMP].append(metadata__timestamp)
        df_columns[Events.META_UPDATED_ON].append(metadata__updated_on)
        df_columns[Events.META_ENRICHED_ON].append(metadata__enriched_on)

        # If called after '__add_sh_info', item will already contain
        # 'grimoire_creation_date'

        if Events.GRIMOIRE_CREATION_DATE in item:
            creation_date = item[Events.GRIMOIRE_CREATION_DATE]
        else:
            creation_date = str_to_datetime(item['data']['AuthorDate'])

        df_columns[Events.GRIMOIRE_CREATION_DATE].append(creation_date)

        # Perceval fields
        df_columns[Events.PERCEVAL_UUID].append(item['uuid'])

        # TODO add other common fields as 'perceval version', 'tag', 'origin'...

    def _add_general_info(self, df_columns, item):

        project_item = self.enrich.get_item_project(item)
        df_columns[Events.PROJECT].append(project_item[Events.PROJECT])
        df_columns[Events.PROJECT_1].append(project_item[Events.PROJECT_1])

    def _add_sh_info(self, df_columns, item):
        # Add the grimoire_creation_date to the raw item
        # It is used for getting the right affiliation
        item.update(self.enrich.get_grimoire_fields(
            item["data"]["AuthorDate"], "commit"))
        sh_identity = self.enrich.get_item_sh(item)

        author_id = sh_identity.get(Events.SH_AUTHOR_ID, Events.UNKNOWN)
        author_org_name = sh_identity.get(Events.SH_AUTHOR_ORG_NAME, Events.UNKNOWN)
        author_name = sh_identity.get(Events.SH_AUTHOR_NAME, Events.UNKNOWN)
        author_uuid = sh_identity.get(Events.SH_AUTHOR_UUID, Events.UNKNOWN)
        author_domain = sh_identity.get(Events.SH_AUTHOR_DOMAIN, Events.UNKNOWN)
        author_username = sh_identity.get(Events.SH_AUTHOR_USER_NAME, Events.UNKNOWN)
        author_bot = sh_identity.get(Events.SH_AUTHOR_BOT, False)
        author_multi_org_names = sh_identity.get(Events.SH_AUTHOR_MULTI_ORG_NAMES, [Events.UNKNOWN])

        df_columns[Events.SH_AUTHOR_ID].append(author_id)
        df_columns[Events.SH_AUTHOR_ORG_NAME].append(author_org_name)
        df_columns[Events.SH_AUTHOR_NAME].append(author_name)
        df_columns[Events.SH_AUTHOR_UUID].append(author_uuid)
        df_columns[Events.SH_AUTHOR_DOMAIN].append(author_domain)
        df_columns[Events.SH_AUTHOR_USER_NAME].append(author_username)
        df_columns[Events.SH_AUTHOR_BOT].append(author_bot)
        df_columns[Events.SH_AUTHOR_MULTI_ORG_NAMES].append(author_multi_org_names)

    def _init_common_fields(self, df_columns):
        # Metadata fields
        df_columns[Events.META_TIMESTAMP] = []
        df_columns[Events.META_UPDATED_ON] = []
        df_columns[Events.META_ENRICHED_ON] = []

        # Common fields
        df_columns[Events.GRIMOIRE_CREATION_DATE] = []
        df_columns[Events.PROJECT] = []
        df_columns[Events.PROJECT_1] = []
        df_columns[Events.PERCEVAL_UUID] = []

        # SortigHat information
        df_columns[Events.SH_AUTHOR_ID] = []
        df_columns[Events.SH_AUTHOR_ORG_NAME] = []
        df_columns[Events.SH_AUTHOR_NAME] = []
        df_columns[Events.SH_AUTHOR_UUID] = []
        df_columns[Events.SH_AUTHOR_DOMAIN] = []
        df_columns[Events.SH_AUTHOR_USER_NAME] = []
        df_columns[Events.SH_AUTHOR_BOT] = []
        df_columns[Events.SH_AUTHOR_MULTI_ORG_NAMES] = []

    def _add_common_fields(self, df_columns, item):
        self._add_metadata(df_columns, item)
        self._add_sh_info(df_columns, item)
        self._add_general_info(df_columns, item)

    @staticmethod
    def _add_common_events(events, df_columns):
        events[Events.META_TIMESTAMP] = df_columns[Events.META_TIMESTAMP]
        events[Events.META_UPDATED_ON] = df_columns[Events.META_UPDATED_ON]
        events[Events.META_ENRICHED_ON] = df_columns[Events.META_ENRICHED_ON]

        events[Events.GRIMOIRE_CREATION_DATE] = df_columns[Events.GRIMOIRE_CREATION_DATE]
        events[Events.PROJECT] = df_columns[Events.PROJECT]
        events[Events.PROJECT_1] = df_columns[Events.PROJECT_1]

        events[Events.PERCEVAL_UUID] = df_columns[Events.PERCEVAL_UUID]

        events[Events.SH_AUTHOR_ID] = df_columns[Events.SH_AUTHOR_ID]
        events[Events.SH_AUTHOR_ORG_NAME] = df_columns[Events.SH_AUTHOR_ORG_NAME]
        events[Events.SH_AUTHOR_NAME] = df_columns[Events.SH_AUTHOR_NAME]
        events[Events.SH_AUTHOR_UUID] = df_columns[Events.SH_AUTHOR_UUID]
        events[Events.SH_AUTHOR_DOMAIN] = df_columns[Events.SH_AUTHOR_DOMAIN]
        events[Events.SH_AUTHOR_USER_NAME] = df_columns[Events.SH_AUTHOR_USER_NAME]
        events[Events.SH_AUTHOR_BOT] = df_columns[Events.SH_AUTHOR_BOT]
        events[Events.SH_AUTHOR_MULTI_ORG_NAMES] = df_columns[Events.SH_AUTHOR_MULTI_ORG_NAMES]


class Bugzilla(Events):
    """ Class used to 'eventize' Bugzilla items

    This splits information of each item based on a pre-configured mapping.
    There are several levels of events. These levels were created as a way
    to have more or less time consuming generation of events.
    """

    EVENT_OPEN = "ISSUE_OPEN"

    # Fields supported by this module (when a DataFrame is returned)
    ISSUE_ID = "id"
    ISSUE_EVENT = "eventtype"
    ISSUE_DATE = "date"
    ISSUE_OWNER = "owner"

    def __bug_photo(self, item):
        """ Retrieves basic information about the current status of the bug

        That current status contains the photo of the bug at the moment of
        the analysis. These fields will be used later for create extra
        events
        """

    def __init__(self, items):
        """ Main constructor of the class

        :param items: original list of JSON that contains all info about a bug
        :type items: list
        """

        self.items = items

    def eventize(self, granularity):
        """ This splits the JSON information found at self.events into the
        several events. For this there are three different levels of time
        consuming actions: 1-soft, 2-medium and 3-hard.

        Level 1 provides events about open and closed issues.
        Level 2 provides events about the rest of the status updates.
        Level 3 provides events about the rest of the values in any of the
        fields.

        :param granularity: Levels of time consuming actions to calculate events
        :type granularity: integer

        :returns: Pandas dataframe with splitted events.
        :rtype: pandas.DataFrame
        """

        issue = {}
        issue[Bugzilla.ISSUE_ID] = []
        issue[Bugzilla.ISSUE_EVENT] = []
        issue[Bugzilla.ISSUE_DATE] = []
        issue[Bugzilla.ISSUE_OWNER] = []

        events = pandas.DataFrame()

        for item in self.items:
            bug_data = item["data"]
            if granularity == 1:
                # Open Date: filling a new event
                issue[Bugzilla.ISSUE_ID].append(bug_data['bug_id'][0]['__text__'])
                issue[Bugzilla.ISSUE_EVENT].append(Bugzilla.EVENT_OPEN)
                issue[Bugzilla.ISSUE_DATE].append(str_to_datetime(bug_data['creation_ts'][0]['__text__']))
                issue[Bugzilla.ISSUE_OWNER].append(bug_data['reporter'][0]["__text__"])

                # Adding the rest of the status updates (if there were any)
                if 'activity' in bug_data.keys():
                    activity = bug_data["activity"]
                    for change in activity:
                        # if change["What"] == "Status":
                        # Filling a new event
                        issue[Bugzilla.ISSUE_ID].append(bug_data['bug_id'][0]['__text__'])
                        issue[Bugzilla.ISSUE_EVENT].append("ISSUE_" + change["Added"])
                        issue[Bugzilla.ISSUE_DATE].append(str_to_datetime(change["When"]))
                        issue[Bugzilla.ISSUE_OWNER].append(change["Who"])

            if granularity == 2:
                # TBD Let's produce an index with all of the changes.
                #    Let's have in mind the point about having the changes of initiating
                #    the ticket.
                pass

            if granularity == 3:
                # TDB
                pass

        # Done in this way to have an order (and not a direct cast)
        events[Bugzilla.ISSUE_ID] = issue[Bugzilla.ISSUE_ID]
        events[Bugzilla.ISSUE_EVENT] = issue[Bugzilla.ISSUE_EVENT]
        events[Bugzilla.ISSUE_DATE] = issue[Bugzilla.ISSUE_DATE]
        events[Bugzilla.ISSUE_OWNER] = issue[Bugzilla.ISSUE_OWNER]

        return events


class BugzillaRest(Events):
    """ Class used to eventize Bugzilla Rest items

    This splits each item information based on a pre-existing mapping.
    """

    EVENT_OPEN = "ISSUE_OPEN"

    # Fields supported by this module (when a DataFrame is returned)
    ISSUE_ID = "id"
    ISSUE_EVENT = "eventtype"
    ISSUE_DATE = "date"
    ISSUE_OWNER = "owner"
    ISSUE_ADDED = "added"
    ISSUE_REMOVED = "removed"

    def __init__(self, items):
        """ Main constructor of the class

        :param items: original list of JSON that contains all info about a bug
        :type items: list
        """

        self.items = items

    def eventize(self, granularity):
        """ This splits the JSON information found at self.events into the
        several events. For this there are three different levels of time
        consuming actions: 1-soft, 2-medium and 3-hard.

        Level 1 provides events about open and closed issues.
        Level 2 provides events about the rest of the status updates.
        Level 3 provides events about the rest of the values in any of the
        fields.

        :param granularity: Levels of time consuming actions to calculate events
        :type granularity: integer

        :returns: Pandas dataframe with splitted events.
        :rtype: pandas.DataFrame
        """

        issue = {}
        issue[BugzillaRest.ISSUE_ID] = []
        issue[BugzillaRest.ISSUE_EVENT] = []
        issue[BugzillaRest.ISSUE_DATE] = []
        issue[BugzillaRest.ISSUE_OWNER] = []
        issue[BugzillaRest.ISSUE_ADDED] = []
        issue[BugzillaRest.ISSUE_REMOVED] = []

        events = pandas.DataFrame()

        for item in self.items:
            bug_data = item["data"]
            if granularity == 1:
                # Open Date: filling a new event
                issue[BugzillaRest.ISSUE_ID].append(bug_data['id'])
                issue[BugzillaRest.ISSUE_EVENT].append(BugzillaRest.EVENT_OPEN)
                issue[BugzillaRest.ISSUE_DATE].append(str_to_datetime(bug_data['creation_time']))
                issue[BugzillaRest.ISSUE_OWNER].append(bug_data['creator_detail']["real_name"])
                issue[BugzillaRest.ISSUE_ADDED].append("-")
                issue[BugzillaRest.ISSUE_REMOVED].append("-")

                # Adding the rest of the status updates (if there were any)
                if 'history' in bug_data.keys():
                    history = bug_data["history"]
                    for step in history:
                        # Filling a new event
                        who = step["who"]
                        when = str_to_datetime(step["when"])
                        changes = step["changes"]
                        for change in changes:
                            issue[BugzillaRest.ISSUE_ID].append(bug_data['id'])
                            issue[BugzillaRest.ISSUE_EVENT].append("ISSUE_" + change["field_name"])
                            issue[BugzillaRest.ISSUE_ADDED].append(change["added"])
                            issue[BugzillaRest.ISSUE_REMOVED].append(change["removed"])
                            issue[BugzillaRest.ISSUE_DATE].append(when)
                            issue[BugzillaRest.ISSUE_OWNER].append(who)

            if granularity == 2:
                # TBD Let's produce an index with all of the changes.
                #    Let's have in mind the point about having the changes of initiating
                #    the ticket.
                pass

            if granularity == 3:
                # TDB
                pass

        # Done in this way to have an order (and not a direct cast)
        events[BugzillaRest.ISSUE_ID] = issue[BugzillaRest.ISSUE_ID]
        events[BugzillaRest.ISSUE_EVENT] = issue[BugzillaRest.ISSUE_EVENT]
        events[BugzillaRest.ISSUE_DATE] = issue[BugzillaRest.ISSUE_DATE]
        events[BugzillaRest.ISSUE_OWNER] = issue[BugzillaRest.ISSUE_OWNER]
        events[BugzillaRest.ISSUE_ADDED] = issue[BugzillaRest.ISSUE_ADDED]
        events[BugzillaRest.ISSUE_REMOVED] = issue[BugzillaRest.ISSUE_REMOVED]

        return events


class Git(Events):
    """ Class used to 'eventize' Git items

    This splits information of each item based on a pre-configured mapping.
    There are several levels of events. These levels were created as a way
    to have more or less time consuming generation of events.
    """

    EVENT_COMMIT = "COMMIT"
    EVENT_FILE = "FILE_"

    # Fields supported by this module (when a DataFrame is returned)
    COMMIT_ID = "id"
    COMMIT_EVENT = "eventtype"
    COMMIT_DATE = "date"
    COMMIT_OWNER = "owner"
    COMMIT_COMMITTER = "committer"
    COMMIT_COMMITTER_DATE = "committer_date"
    COMMIT_COMMITTER_TZ = "tz"
    COMMIT_REPOSITORY = "repository"
    COMMIT_MESSAGE = "message"
    COMMIT_NUM_FILES = "num_files"
    COMMIT_ADDED_LINES = "num_added_lines"
    COMMIT_REMOVED_LINES = "num_removed_lines"

    COMMIT_HASH = "hash"

    AUTHOR_DOMAIN = "git_author_domain"

    FILE_EVENT = "fileaction"
    FILE_PATH = "filepath"
    FILE_ADDED_LINES = "addedlines"
    FILE_REMOVED_LINES = "removedlines"
    FILE_FILES = "files"

    def __init__(self, items, git_enrich):
        """ Main constructor of the class

        :param items: original list of JSON that contains all info about a commit
        :type items: list
        :param git_enrich:
        :type enrich: grimoire_elk.elk.git.GitEnrich
        """

        super().__init__(items=items, enrich=git_enrich)

    def __add_commit_info(self, df_columns, item):

        commit_data = item["data"]
        repository = item["origin"]

        creation_date = str_to_datetime(commit_data['AuthorDate'])

        df_columns[Git.COMMIT_HASH].append(commit_data['commit'])

        df_columns[Git.COMMIT_ID].append(commit_data['commit'])
        df_columns[Git.COMMIT_EVENT].append(Git.EVENT_COMMIT)
        df_columns[Git.COMMIT_DATE].append(creation_date)
        df_columns[Git.COMMIT_OWNER].append(commit_data['Author'])
        df_columns[Git.COMMIT_COMMITTER].append(commit_data['Commit'])
        df_columns[Git.COMMIT_COMMITTER_DATE].append(str_to_datetime(commit_data['CommitDate']))
        df_columns[Git.COMMIT_REPOSITORY].append(repository)
        if 'message' in commit_data.keys():
            df_columns[Git.COMMIT_MESSAGE].append(commit_data['message'])
        else:
            df_columns[Git.COMMIT_MESSAGE].append('')

        author_domain = self.enrich.get_identity_domain(self.enrich.get_sh_identity(item, 'Author'))
        df_columns[Git.AUTHOR_DOMAIN].append(author_domain)

        try:
            commit_date = str_to_datetime(commit_data['CommitDate'])
            commit_tz = int(commit_date.strftime("%z")[0:3])
        except ValueError:
            commit_tz = 0
        df_columns[Git.COMMIT_COMMITTER_TZ].append(commit_tz)

    def eventize(self, granularity):
        """ This splits the JSON information found at self.events into the
        several events. For this there are three different levels of time
        consuming actions: 1-soft, 2-medium and 3-hard.

        Level 1 provides events about commits
        Level 2 provides events about files
        Level 3 provides other events (not used so far)

        :param granularity: Levels of time consuming actions to calculate events
        :type granularity: integer

        :returns: Pandas dataframe with splitted events.
        :rtype: pandas.DataFrame
        """

        df_columns = {}
        # Init common columns
        self._init_common_fields(df_columns)

        # First level granularity
        df_columns[Git.COMMIT_ID] = []
        df_columns[Git.COMMIT_EVENT] = []
        df_columns[Git.COMMIT_DATE] = []
        df_columns[Git.COMMIT_OWNER] = []
        df_columns[Git.COMMIT_COMMITTER] = []
        df_columns[Git.COMMIT_COMMITTER_DATE] = []
        df_columns[Git.COMMIT_COMMITTER_TZ] = []
        df_columns[Git.COMMIT_REPOSITORY] = []
        df_columns[Git.COMMIT_MESSAGE] = []
        df_columns[Git.COMMIT_NUM_FILES] = []
        df_columns[Git.COMMIT_ADDED_LINES] = []
        df_columns[Git.COMMIT_REMOVED_LINES] = []
        df_columns[Git.COMMIT_HASH] = []
        df_columns[Git.AUTHOR_DOMAIN] = []

        # Second level of granularity
        df_columns[Git.FILE_FILES] = []
        df_columns[Git.FILE_EVENT] = []
        df_columns[Git.FILE_PATH] = []
        df_columns[Git.FILE_ADDED_LINES] = []
        df_columns[Git.FILE_REMOVED_LINES] = []

        events = pandas.DataFrame()

        for item in self.items:
            commit_data = item["data"]
            if granularity == 1:
                self._add_common_fields(df_columns, item)
                self.__add_commit_info(df_columns, item)

                added_lines = 0
                removed_lines = 0
                files = commit_data["files"]
                df_columns[Git.COMMIT_NUM_FILES] = int(len(files))
                for f in files:
                    if "added" in f.keys() and f["added"] != "-":
                        added_lines = added_lines + int(f["added"])
                    if "removed" in f.keys() and f["removed"] != "-":
                        removed_lines = removed_lines + int(f["removed"])
                df_columns[Git.COMMIT_ADDED_LINES] = added_lines
                df_columns[Git.COMMIT_REMOVED_LINES] = removed_lines

            # TODO: this will fail if no files are found in a commit (eg: merge)
            if granularity == 2:
                # Add extra info about files actions, if there were any
                if "files" in commit_data.keys():
                    files = commit_data["files"]
                    nfiles = 0
                    for f in files:
                        if "action" in f.keys():
                            nfiles += 1

                    for f in files:
                        self._add_common_fields(df_columns, item)
                        self.__add_commit_info(df_columns, item)

                        df_columns[Git.FILE_FILES].append(nfiles)

                        if "action" in f.keys():
                            df_columns[Git.FILE_EVENT].append(Git.EVENT_FILE + f["action"])
                        else:
                            df_columns[Git.FILE_EVENT].append("-")

                        if "file" in f.keys():
                            df_columns[Git.FILE_PATH].append(f["file"])
                        else:
                            df_columns[Git.FILE_PATH].append("-")

                        if "added" in f.keys():
                            if f["added"] == "-":
                                df_columns[Git.FILE_ADDED_LINES].append(0)
                            else:
                                df_columns[Git.FILE_ADDED_LINES].append(int(f["added"]))
                        else:
                            df_columns[Git.FILE_ADDED_LINES].append(0)

                        if "removed" in f.keys():
                            if f["removed"] == "-":
                                df_columns[Git.FILE_REMOVED_LINES].append(0)
                            else:
                                df_columns[Git.FILE_REMOVED_LINES].append(int(f["removed"]))
                        else:
                            df_columns[Git.FILE_REMOVED_LINES].append(0)

                else:
                    print("Merge found, doing nothing...")

            if granularity == 3:
                # TDB
                pass

        # Done in this way to have an order (and not a direct cast)
        self._add_common_events(events, df_columns)

        events[Git.COMMIT_ID] = df_columns[Git.COMMIT_ID]
        events[Git.COMMIT_EVENT] = df_columns[Git.COMMIT_EVENT]
        events[Git.COMMIT_DATE] = df_columns[Git.COMMIT_DATE]
        events[Git.COMMIT_OWNER] = df_columns[Git.COMMIT_OWNER]
        events[Git.COMMIT_COMMITTER] = df_columns[Git.COMMIT_COMMITTER]
        events[Git.COMMIT_COMMITTER_DATE] = df_columns[Git.COMMIT_COMMITTER_DATE]
        events[Git.COMMIT_COMMITTER_TZ] = df_columns[Git.COMMIT_COMMITTER_TZ]
        events[Git.COMMIT_REPOSITORY] = df_columns[Git.COMMIT_REPOSITORY]
        events[Git.COMMIT_MESSAGE] = df_columns[Git.COMMIT_MESSAGE]
        events[Git.COMMIT_HASH] = df_columns[Git.COMMIT_HASH]
        events[Git.AUTHOR_DOMAIN] = df_columns[Git.AUTHOR_DOMAIN]

        if granularity == 1:
            events[Git.COMMIT_NUM_FILES] = df_columns[Git.COMMIT_NUM_FILES]
            events[Git.COMMIT_ADDED_LINES] = df_columns[Git.COMMIT_ADDED_LINES]
            events[Git.COMMIT_REMOVED_LINES] = df_columns[Git.COMMIT_REMOVED_LINES]
        if granularity == 2:
            events[Git.FILE_FILES] = df_columns[Git.FILE_FILES]
            events[Git.FILE_EVENT] = df_columns[Git.FILE_EVENT]
            events[Git.FILE_PATH] = df_columns[Git.FILE_PATH]
            events[Git.FILE_ADDED_LINES] = df_columns[Git.FILE_ADDED_LINES]
            events[Git.FILE_REMOVED_LINES] = df_columns[Git.FILE_REMOVED_LINES]

        return events


class Gerrit(Events):
    """ Class used to 'eventize' Gerrit items

    This splits information of each item based on a pre-configured mapping.
    There are several levels of events. These levels were created as a way
    to have more or less time consuming generation of events.
    """

    EVENT_OPEN = "CHANGESET_SENT"
    EVENT_ = "CHANGESET_"

    # Fields supported by this module (when a DataFrame is returned)
    CHANGESET_ID = "id"
    CHANGESET_EVENT = "eventtype"
    CHANGESET_DATE = "date"
    CHANGESET_OWNER = "owner"
    CHANGESET_EMAIL = "email"
    CHANGESET_VALUE = "value"
    CHANGESET_REPO = "repository"

    def __init__(self, items):
        """ Main constructor of the class

        :param items: original list of JSON that contains all info about a commit
        :type items: list
        """

        self.items = items

    def eventize(self, granularity):
        """ This splits the JSON information found at self.events into the
        several events. For this there are three different levels of time
        consuming actions: 1-soft, 2-medium and 3-hard.

        Level 1 provides events about commits
        Level 2 provides events about files
        Level 3 provides other events (not used so far)

        :param granularity: Levels of time consuming actions to calculate events
        :type granularity: integer

        :returns: Pandas dataframe with splitted events.
        :rtype: pandas.DataFrame
        """

        changeset = {}
        # First level granularity
        changeset[Gerrit.CHANGESET_ID] = []
        changeset[Gerrit.CHANGESET_EVENT] = []
        changeset[Gerrit.CHANGESET_DATE] = []
        changeset[Gerrit.CHANGESET_OWNER] = []
        changeset[Gerrit.CHANGESET_EMAIL] = []
        changeset[Gerrit.CHANGESET_VALUE] = []
        changeset[Gerrit.CHANGESET_REPO] = []

        events = pandas.DataFrame()

        for item in self.items:
            changeset_data = item["data"]
            if granularity >= 1:
                # Changeset submission date: filling a new event
                changeset[Gerrit.CHANGESET_ID].append(changeset_data["number"])
                changeset[Gerrit.CHANGESET_EVENT].append(Gerrit.EVENT_OPEN)
                changeset[Gerrit.CHANGESET_DATE].append(dt.fromtimestamp(int(changeset_data["createdOn"])))
                changeset[Gerrit.CHANGESET_REPO].append(changeset_data["project"])
                value = email = "notknown"
                if "name" in changeset_data["owner"].keys():
                    value = changeset_data["owner"]["name"]
                elif "username" in changeset_data["owner"].keys():
                    value = changeset_data["owner"]["username"]
                elif "email" in changeset_data["owner"].keys():
                    value = changeset_data["owner"]["email"]
                    email = changeset_data["owner"]["email"]
                changeset[Gerrit.CHANGESET_OWNER].append(value)
                changeset[Gerrit.CHANGESET_EMAIL].append(email)
                changeset[Gerrit.CHANGESET_VALUE].append(-10)

                # Adding the closing status updates (if there was any)
                if changeset_data["status"] == 'ABANDONED' or \
                   changeset_data["status"] == 'MERGED':
                    closing_date = dt.fromtimestamp(int(changeset_data["lastUpdated"]))
                    changeset[Gerrit.CHANGESET_ID].append(changeset_data["number"])
                    changeset[Gerrit.CHANGESET_EVENT].append(Gerrit.EVENT_ +
                                                             changeset_data["status"])
                    changeset[Gerrit.CHANGESET_DATE].append(closing_date)
                    changeset[Gerrit.CHANGESET_REPO].append(changeset_data["project"])
                    value = email = "notknown"
                    if "name" in changeset_data["owner"].keys():
                        value = changeset_data["owner"]["name"]
                    if "username" in changeset_data["owner"].keys():
                        value = changeset_data["owner"]["username"]
                    if "email" in changeset_data["owner"].keys():
                        value = changeset_data["owner"]["email"]
                        email = changeset_data["owner"]["email"]
                    changeset[Gerrit.CHANGESET_OWNER].append(value)
                    changeset[Gerrit.CHANGESET_EMAIL].append(email)
                    changeset[Gerrit.CHANGESET_VALUE].append(-10)

            if granularity >= 2:
                # Adding extra info about the patchsets
                for patchset in changeset_data["patchSets"]:
                    changeset[Gerrit.CHANGESET_ID].append(changeset_data["number"])
                    changeset[Gerrit.CHANGESET_EVENT].append(Gerrit.EVENT_ + "PATCHSET_SENT")
                    changeset[Gerrit.CHANGESET_DATE].append(
                        dt.fromtimestamp(int(patchset["createdOn"])))
                    changeset[Gerrit.CHANGESET_REPO].append(changeset_data["project"])
                    try:
                        email = "patchset_noname"
                        if "name" in patchset["author"].keys():
                            value = patchset["author"]["name"]
                        if "username" in patchset["author"].keys():
                            value = patchset["author"]["username"]
                        if "email" in patchset["author"].keys():
                            value = patchset["author"]["email"]
                            email = patchset["author"]["email"]
                    except KeyError:
                        value = "patchset_noname"
                    changeset[Gerrit.CHANGESET_OWNER].append(value)
                    changeset[Gerrit.CHANGESET_EMAIL].append(email)
                    changeset[Gerrit.CHANGESET_VALUE].append(-10)
                    # print (patchset)
                    if "approvals" in patchset.keys():
                        for approval in patchset["approvals"]:
                            if approval["type"] != "Code-Review":
                                continue
                            changeset[Gerrit.CHANGESET_ID].append(changeset_data["number"])
                            changeset[Gerrit.CHANGESET_EVENT].append(
                                Gerrit.EVENT_ +
                                "PATCHSET_APPROVAL_" + approval["type"])
                            changeset[Gerrit.CHANGESET_DATE].append(
                                dt.fromtimestamp(int(approval["grantedOn"])))
                            changeset[Gerrit.CHANGESET_REPO].append(changeset_data["project"])
                            email = "approval_noname"
                            if "name" in approval["by"].keys():
                                value = approval["by"]["name"]
                            elif "username" in approval["by"].keys():
                                value = approval["by"]["username"]
                            elif "email" in approval["by"].keys():
                                value = approval["by"]["email"]
                                email = approval["by"]["email"]
                            changeset[Gerrit.CHANGESET_OWNER].append(value)
                            changeset[Gerrit.CHANGESET_EMAIL].append(email)
                            changeset[Gerrit.CHANGESET_VALUE].append(int(approval["value"]))

            if granularity >= 3:
                # TDB
                pass

        # Done in this way to have an order (and not a direct cast)
        events[Gerrit.CHANGESET_ID] = changeset[Gerrit.CHANGESET_ID]
        events[Gerrit.CHANGESET_EVENT] = changeset[Gerrit.CHANGESET_EVENT]
        events[Gerrit.CHANGESET_DATE] = changeset[Gerrit.CHANGESET_DATE]
        events[Gerrit.CHANGESET_OWNER] = changeset[Gerrit.CHANGESET_OWNER]
        events[Gerrit.CHANGESET_EMAIL] = changeset[Gerrit.CHANGESET_EMAIL]
        events[Gerrit.CHANGESET_VALUE] = changeset[Gerrit.CHANGESET_VALUE]
        events[Gerrit.CHANGESET_REPO] = changeset[Gerrit.CHANGESET_REPO]

        return events


class Email(Events):
    """ Class used to 'eventize' mailing list items

    This splits information of each item based on a pre-configured mapping.
    There are several levels of events. These levels were created as a way
    to have more or less time consuming generation of events.
    """

    EVENT_OPEN = "EMAIL_SENT"

    # Fields supported by this module (when a DataFrame is returned)
    EMAIL_ID = "id"
    EMAIL_EVENT = "eventtype"
    EMAIL_DATE = "date"
    EMAIL_OWNER = "owner"
    EMAIL_SUBJECT = "subject"
    EMAIL_BODY = "body"
    EMAIL_ORIGIN = "mailinglist"

    def __init__(self, items):
        """ Main constructor of the class

        :param items: original list of JSON that contains all info about a commit
        :type items: list
        """

        self.items = items

    def eventize(self, granularity):
        """ This splits the JSON information found at self.events into the
        several events. For this there are three different levels of time
        consuming actions: 1-soft, 2-medium and 3-hard.

        Level 1 provides events about emails
        Level 2 not implemented
        Level 3 not implemented

        :param granularity: Levels of time consuming actions to calculate events
        :type granularity: integer

        :returns: Pandas dataframe with splitted events.
        :rtype: pandas.DataFrame
        """

        email = {}
        # First level granularity
        email[Email.EMAIL_ID] = []
        email[Email.EMAIL_EVENT] = []
        email[Email.EMAIL_DATE] = []
        email[Email.EMAIL_OWNER] = []
        email[Email.EMAIL_SUBJECT] = []
        email[Email.EMAIL_BODY] = []
        email[Email.EMAIL_ORIGIN] = []

        events = pandas.DataFrame()

        for item in self.items:
            origin = item["origin"]
            email_data = item["data"]
            if granularity == 1:
                # Changeset submission date: filling a new event
                email[Email.EMAIL_ID].append(email_data["Message-ID"])
                email[Email.EMAIL_EVENT].append(Email.EVENT_OPEN)
                try:
                    email[Email.EMAIL_DATE].append(str_to_datetime(email_data["Date"], ignoretz=True))
                except KeyError:
                    email[Email.EMAIL_DATE].append(str_to_datetime("1970-01-01"))
                email[Email.EMAIL_OWNER].append(email_data["From"])
                email[Email.EMAIL_SUBJECT].append(email_data["Subject"])
                try:
                    email[Email.EMAIL_BODY].append(email_data["body"]["plain"])
                except KeyError:
                    email[Email.EMAIL_BODY].append("None")
                email[Email.EMAIL_ORIGIN].append(origin)

            if granularity == 2:
                # TDB
                pass

            if granularity == 3:
                # TDB
                pass

        # Done in this way to have an order (and not a direct cast)
        events[Email.EMAIL_ID] = email[Email.EMAIL_ID]
        events[Email.EMAIL_EVENT] = email[Email.EMAIL_EVENT]
        events[Email.EMAIL_DATE] = email[Email.EMAIL_DATE]
        events[Email.EMAIL_OWNER] = email[Email.EMAIL_OWNER]
        events[Email.EMAIL_SUBJECT] = email[Email.EMAIL_SUBJECT]
        events[Email.EMAIL_BODY] = email[Email.EMAIL_BODY]
        events[Email.EMAIL_ORIGIN] = email[Email.EMAIL_ORIGIN]

        return events



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2019 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#     Daniel Izquierdo <dizquierdo@bitergia.com>
#     Alberto Perez <alpgarcia@bitergia.com>
#

import os
import sys
import unittest

if '..' not in sys.path:
    sys.path.insert(0, '..')

import json

from cereslib.events.events import Git

from grimoire_elk.utils import get_connectors


class TestEvents(unittest.TestCase):
    """ Unit tests for Enrich classes
    """

    def setUp(self):
        self.__tests_dir = os.path.dirname(os.path.realpath(__file__))
        self.__events_dir = os.path.join(self.__tests_dir, "data/events/")
        self.connectors = get_connectors()

    def test_GitEvents(self):
        """ Test several cases for the PairProgramming class
        """

        connector = "git"

        with open(os.path.join(self.__events_dir, connector + ".json")) as f:
            items = json.load(f)

        enrich_backend = self.connectors[connector][2](json_projects_map="./data/projects_map.json")
        enrich_backend.sortinghat = True
        events = Git(items, enrich_backend)

        events_df = events.eventize(1)

        self.assertFalse(events_df.empty)
        self.assertEqual(len(events_df), 6)
        self.assertEqual(len(events_df.columns), 28)
        self.assertIn("metadata__timestamp", events_df)
        self.assertIn("metadata__updated_on", events_df)
        self.assertIn("metadata__enriched_on", events_df)
        self.assertIn("grimoire_creation_date", events_df)
        self.assertIn("project", events_df)
        self.assertIn("project_1", events_df)
        self.assertIn("perceval_uuid", events_df)
        self.assertIn("author_id", events_df)
        self.assertIn("author_org_name", events_df)
        self.assertIn("author_name", events_df)
        self.assertIn("author_uuid", events_df)
        self.assertIn("author_domain", events_df)
        self.assertIn("author_user_name", events_df)
        self.assertIn("author_bot", events_df)
        self.assertIn("author_multi_org_names", events_df)
        self.assertIn("id", events_df)
        self.assertIn("date", events_df)
        self.assertIn("owner", events_df)
        self.assertIn("committer", events_df)
        self.assertIn("committer_date", events_df)
        self.assertIn("tz", events_df)
        self.assertIn("repository", events_df)
        self.assertIn("message", events_df)
        self.assertIn("hash", events_df)
        self.assertIn("git_author_domain", events_df)

        events_df = events.eventize(2)

        self.assertFalse(events_df.empty)
        self.assertEqual(len(events_df), 10)
        self.assertEqual(len(events_df.columns), 30)
        self.assertIn("eventtype", events_df)
        self.assertIn("date", events_df)
        self.assertIn("owner", events_df)
        self.assertIn("committer", events_df)
        self.assertIn("committer_date", events_df)
        self.assertIn("tz", events_df)
        self.assertIn("repository", events_df)
        self.assertIn("message", events_df)
        self.assertIn("hash", events_df)
        self.assertIn("git_author_domain", events_df)
        self.assertIn("files", events_df)
        self.assertIn("fileaction", events_df)
        self.assertIn("filepath", events_df)
        self.assertIn("addedlines", events_df)
        self.assertIn("removedlines", events_df)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2019 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#     Alberto Pérez García-Plaza <alpgarcia@bitergia.com>
#

import sys
import unittest

import pandas

if '..' not in sys.path:
    sys.path.insert(0, '..')

from cereslib.dfutils.filter import FilterRows


class TestFilter(unittest.TestCase):
    """ Unit tests for Filter class
    """

    def test_filter_rows(self):
        """ Test several cases for filtering rows by column value
        """

        # One column, values of different types
        df = pandas.DataFrame()
        filepaths = ['', None, '-', '/file/path', 1, True, pandas.NA, '-', [1, 2]]
        df["filepath"] = filepaths
        data_filtered = FilterRows(df)
        df = data_filtered.filter_(["filepath"], "-")

        self.assertEqual(len(df), 7)

        # One empty column
        df = pandas.DataFrame()
        df["filepath"] = []
        data_filtered = FilterRows(df)
        df = data_filtered.filter_(["filepath"], "-")

        self.assertEqual(len(df), 0)

        # Several columns and just one empty
        df = pandas.DataFrame()
        df["filepath"] = []
        df["name"] = ["name", "-", "other", "-"]
        df["dirname"] = ["dir", "-", "-", "-"]
        data_filtered = FilterRows(df)
        df = data_filtered.filter_(["filepath", "name", "dirname"], "-")

        self.assertEqual(len(df), 1)

    def test_column_not_exists(self):
        """ Test empty dataframe looking for the corresponding ValueError exception
        """
        df = pandas.DataFrame()
        data_filtered = FilterRows(df)
        with self.assertRaisesRegex(ValueError, "Column filepath not in DataFrame columns: \[\]") as context:
            data_filtered.filter_(["filepath"], "-")


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2019 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Daniel Izquierdo Cortazar <dizquierdo@bitergia.com>
#


class Filter(object):
    """ Class that filters information for a given dataset.

    Given a dataframe, there are cases where some information is
    not needed any more. Performance purposes or smaller datasets
    could be part of the reason for this type of needs.

    Thus, this would return a subset of the rows of a given
    dataset as a result of the filtering process.

    An example of this filtering process would be the removal of
    certain commits that follow some pattern such as merges
    or automatically generated ones.
    """


class FilterRows(Filter):
    """ Class used to filter those files that are part of a merge
    Those are initially detected as having no file or no action
    in the list of attributes. This class removes those lines.
    """

    def __init__(self, data):
        """ Main constructor of the class

        :param data: Data frame to be filtered
        :type data: pandas.DataFrame
        """

        self.data = data

    def filter_(self, columns, value):
        """ This method filter some of the rows where the 'value'
        is found in each of the 'columns'.

        :param column: list of strings
        :param value: any type

        :returns: filtered dataframe
        :rtype: pandas.DataFrame
        """

        for column in columns:
            if column not in self.data.columns:
                raise ValueError("Column %s not in DataFrame columns: %s" % (column, list(self.data)))

        for column in columns:
            # Filtering on empty data series doesn't make sense at all and also would raise an error
            column_len = len(self.data[column])
            if column_len > 0 and column_len != self.data[column].isnull().sum():
                self.data = self.data[self.data[column] != value]

        return self.data



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2019 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#     Daniel Izquierdo <dizquierdo@bitergia.com>
#

import sys
import unittest

import pandas

import numpy as np

if '..' not in sys.path:
    sys.path.insert(0, '..')

from cereslib.dfutils.format import Format


class TestFormat(unittest.TestCase):
    """ Unit tests for Format class
    """

    def test_fill_missing_fields(self):
        """ Test several cases for the fill_missing_fields method
        """

        empty_columns = []
        columns = ["test1"]
        empty_df = pandas.DataFrame()
        # With empty dataframe and any columns, this always returns empty dataframe
        # A DataFrame with columns but not data is an empty DataFrame
        self.assertTrue(Format().fill_missing_fields(empty_df, empty_columns).empty)
        self.assertTrue(Format().fill_missing_fields(empty_df, columns).empty)
        self.assertEqual(columns, Format().fill_missing_fields(empty_df, columns).columns)

        # With a dataframe with some data, this returns a non-empty dataframe
        df = empty_df.copy()
        df["test"] = np.zeros(10)

        self.assertFalse(Format().fill_missing_fields(df, empty_columns).empty)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2019 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Daniel Izquierdo Cortazar <dizquierdo@bitergia.com>
#


import pandas

import numpy as np


class Format(object):
    """ Library that allows to format dataframes to be later enriched

    This class is the first step in the enrichment process of data.
    Although this can be used alone for other purposes, its main
    goal consists of providing well formated [missing fields,
    string dates, removal of not needed fields] for the following
    steps of the enrichment process.

    This data format and cleaning process is done due to
    inconsistencies and missing fields that may appear when reading
    information.
    """

    def fill_missing_fields(self, data, columns):
        """ This  method fills with 0's missing fields

        :param data: original Pandas dataframe
        :param columns: list of columns to be filled in the DataFrame
        :type data: pandas.DataFrame
        :type columns: list of strings

        :returns: Pandas dataframe with missing fields filled with 0's
        :rtype: pandas.DataFrame
        """

        for column in columns:
            if column not in data.columns:
                data[column] = np.zeros(len(data))

        return data

    def update_field_names(self, data, matching):
        """ This method updates the names of the fields according to matching

        :param data: original Pandas dataframe
        :param matching: dictionary of matchings between old and new values
        :type data: pandas.DataFrame
        :type matching: dictionary

        :returns: Pandas dataframe with updated names
        :rtype: pandas.DataFrame
        """

        for key in matching.keys():
            if key in data.columns:
                data.rename(columns={key: matching[key]})

        return data

    def format_dates(self, data, columns):
        """ This method translates columns values into datetime objects

        :param data: original Pandas dataframe
        :param columns: list of columns to cast the date to a datetime object
        :type data: pandas.DataFrame
        :type columns: list of strings

        :returns: Pandas dataframe with updated 'columns' with datetime objects
        :rtype: pandas.DataFrame
        """

        for column in columns:
            if column in data.columns:
                data[column] = pandas.to_datetime(data[column])

        return data

    def remove_columns(self, data, columns):
        """ This method removes columns in data

        :param data: original Pandas dataframe
        :param columns: list of columns to remove
        :type data: pandas.DataFrame
        :type columns: list of strings

        :returns: Pandas dataframe with removed columns
        :rtype: pandas.DataFrame
        """

        for column in columns:
            if column in data.columns:
                data = data.drop(column, axis=1)

        return data



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2019 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#     Daniel Izquierdo <dizquierdo@bitergia.com>
#     Alberto Perez <alpgarcia@bitergia.com>
#

import os
import sys
import unittest

import pandas

if '..' not in sys.path:
    sys.path.insert(0, '..')

from cereslib.enrich.enrich import PairProgramming, TimeDifference, Uuid, FilePath
from cereslib.enrich.enrich import Onion

from cereslib.dfutils.format import Format


class TestEnrich(unittest.TestCase):
    """ Unit tests for Enrich classes
    """

    def setUp(self):
        self.__tests_dir = os.path.dirname(os.path.realpath(__file__))
        self.__enrich_dir = os.path.join(self.__tests_dir, "data/enrich/")

    def test_PairProgramming(self):
        """ Test several cases for the PairProgramming class
        """

        empty_df = pandas.DataFrame()
        # With empty dataframe and any columns, this always returns empty dataframe
        # A DataFrame with columns but not data is an empty DataFrame
        pair = PairProgramming(empty_df)
        self.assertTrue(pair.enrich("test1", "test2").empty)
        # With a dataframe with some data, this returns a non-empty dataframe

        csv_df = pandas.read_csv(os.path.join(self.__enrich_dir,
                                              "pairprogramming.csv"))
        pair = PairProgramming(csv_df)
        enriched_df = pair.enrich("test1", "test2")
        self.assertFalse(enriched_df.empty)
        # Expected to return the same df as test1 and test2 do not exist
        self.assertEqual(len(enriched_df), 7)

        enriched_df = pair.enrich("author", "committer")
        self.assertFalse(enriched_df.empty)
        # Expected to add a new entry at the end of the dataframe
        self.assertEqual(len(enriched_df), 8)

        enriched_df = pair.enrich("committer", "author")
        self.assertFalse(enriched_df.empty)
        # Expected to add a new entry at the end of the dataframe
        self.assertEqual(len(enriched_df), 8)

    def test_TimeDifference(self):
        """ Several test cases for the TimeDifference class
        """

        empty_df = pandas.DataFrame()
        # With empty dataframe and any columns this always returns emtpy datafram
        # A DataFrame with columns but no data is an empty DataFrame
        time = TimeDifference(empty_df)
        self.assertTrue(time.enrich("fake_column1", "fake_column2").empty)

        # A dataframe with some data, this returns a non-empty dataframe
        csv_df = pandas.read_csv(os.path.join(self.__enrich_dir,
                                              "timedifference.csv"))
        time = TimeDifference(csv_df)
        format_ = Format()
        time.data = format_.format_dates(time.data, ["date_author", "date_committer"])

        enriched_df = time.enrich("fake_column1", "fake_column2")
        self.assertFalse(enriched_df.empty)
        # Expected to return the same df as fake columns do not exist
        self.assertEqual(len(enriched_df), 7)

        enriched_df = time.enrich("date_author", "date_committer")

        self.assertEqual(len(enriched_df[enriched_df["timedifference"] > 0]), 4)

    def test_FilePath(self):
        """ Test FilePath enricher"""

        # Empty test
        empty_df = pandas.DataFrame()
        filepath = FilePath(empty_df)
        enriched_df = filepath.enrich('filepath')
        self.assertTrue(enriched_df.empty)

        #
        test_df = pandas.DataFrame()
        file_1 = {}
        file_2 = {}
        file_3 = {}
        file_4 = {}
        file_5 = {}
        file_6 = {}
        file_1['filepath'] = 'file.txt'
        file_1['file_name'] = 'file.txt'
        file_1['file_ext'] = 'txt'
        file_1['file_dir_name'] = '/'
        file_1['file_path_list'] = ['file.txt']
        file_2['filepath'] = '/foo/bar'
        file_2['file_name'] = 'bar'
        file_2['file_ext'] = ''
        file_2['file_dir_name'] = '/foo/'
        file_2['file_path_list'] = ['foo', 'bar']
        file_3['filepath'] = '/foo/bar/file.txt'
        file_3['file_name'] = 'file.txt'
        file_3['file_ext'] = 'txt'
        file_3['file_dir_name'] = '/foo/bar/'
        file_3['file_path_list'] = ['foo', 'bar', 'file.txt']
        file_4['filepath'] = '/foo/bar/'
        file_4['file_name'] = ''
        file_4['file_ext'] = ''
        file_4['file_dir_name'] = '/foo/bar/'
        file_4['file_path_list'] = ['foo', 'bar']
        file_5['filepath'] = '/foo//bar.txt'
        file_5['file_name'] = 'bar.txt'
        file_5['file_ext'] = 'txt'
        file_5['file_dir_name'] = '/foo/'
        file_5['file_path_list'] = ['foo', 'bar.txt']
        file_6['filepath'] = '//foo///bar.txt'
        file_6['file_name'] = 'bar.txt'
        file_6['file_ext'] = 'txt'
        file_6['file_dir_name'] = '/foo/'
        file_6['file_path_list'] = ['foo', 'bar.txt']
        test_df['filepath'] = [file_1['filepath'], file_2['filepath'],
                               file_3['filepath'], file_4['filepath'],
                               file_5['filepath'], file_6['filepath']]
        filepath = FilePath(test_df)
        enriched_df = filepath.enrich('filepath')
        self.assertEqual(enriched_df.iloc[[0]]['filepath'].item(), file_1['filepath'])
        self.assertEqual(enriched_df.iloc[[1]]['filepath'].item(), file_2['filepath'])
        self.assertEqual(enriched_df.iloc[[2]]['filepath'].item(), file_3['filepath'])
        self.assertEqual(enriched_df.iloc[[3]]['filepath'].item(), file_4['filepath'])
        self.assertEqual(enriched_df.iloc[[4]]['filepath'].item(), file_5['filepath'])
        self.assertEqual(enriched_df.iloc[[5]]['filepath'].item(), file_6['filepath'])
        self.assertEqual(enriched_df.iloc[[0]]['file_name'].item(), file_1['file_name'])
        self.assertEqual(enriched_df.iloc[[1]]['file_name'].item(), file_2['file_name'])
        self.assertEqual(enriched_df.iloc[[2]]['file_name'].item(), file_3['file_name'])
        self.assertEqual(enriched_df.iloc[[3]]['file_name'].item(), file_4['file_name'])
        self.assertEqual(enriched_df.iloc[[4]]['file_name'].item(), file_5['file_name'])
        self.assertEqual(enriched_df.iloc[[5]]['file_name'].item(), file_6['file_name'])
        self.assertEqual(enriched_df.iloc[[0]]['file_ext'].item(), file_1['file_ext'])
        self.assertEqual(enriched_df.iloc[[1]]['file_ext'].item(), file_2['file_ext'])
        self.assertEqual(enriched_df.iloc[[2]]['file_ext'].item(), file_3['file_ext'])
        self.assertEqual(enriched_df.iloc[[3]]['file_ext'].item(), file_4['file_ext'])
        self.assertEqual(enriched_df.iloc[[4]]['file_ext'].item(), file_5['file_ext'])
        self.assertEqual(enriched_df.iloc[[5]]['file_ext'].item(), file_6['file_ext'])
        self.assertEqual(enriched_df.iloc[[0]]['file_dir_name'].item(), file_1['file_dir_name'])
        self.assertEqual(enriched_df.iloc[[1]]['file_dir_name'].item(), file_2['file_dir_name'])
        self.assertEqual(enriched_df.iloc[[2]]['file_dir_name'].item(), file_3['file_dir_name'])
        self.assertEqual(enriched_df.iloc[[3]]['file_dir_name'].item(), file_4['file_dir_name'])
        self.assertEqual(enriched_df.iloc[[4]]['file_dir_name'].item(), file_5['file_dir_name'])
        self.assertEqual(enriched_df.iloc[[5]]['file_dir_name'].item(), file_6['file_dir_name'])
        self.assertEqual(enriched_df.iloc[[0]]['file_path_list'].item(), file_1['file_path_list'])
        self.assertEqual(enriched_df.iloc[[1]]['file_path_list'].item(), file_2['file_path_list'])
        self.assertEqual(enriched_df.iloc[[2]]['file_path_list'].item(), file_3['file_path_list'])
        self.assertEqual(enriched_df.iloc[[3]]['file_path_list'].item(), file_4['file_path_list'])
        self.assertEqual(enriched_df.iloc[[4]]['file_path_list'].item(), file_5['file_path_list'])
        self.assertEqual(enriched_df.iloc[[5]]['file_path_list'].item(), file_6['file_path_list'])

    def test_Uuid(self):
        """ Test several cases for the Uuid class
        """

        # Empty LEFT dataframe
        empty_df = pandas.DataFrame()
        uuid = Uuid(empty_df,
                    file_path=os.path.join(self.__enrich_dir, "uuids.csv"))

        enriched_df = uuid.enrich(['name', 'email'])

        # If LEFT df is empty, nothing should be merged resulting in an empty df
        self.assertTrue(enriched_df.empty)

        # Load test data from CSV files
        authors_df = pandas.read_csv(os.path.join(self.__enrich_dir,
                                                  "authors.csv"))
        uuid = Uuid(authors_df,
                    file_path=os.path.join(self.__enrich_dir, "uuids.csv"))

        enriched_df = uuid.enrich(['name', 'email'])

        self.assertFalse(enriched_df.empty)
        # Merged len must be equal to LEFT df len
        self.assertEqual(len(enriched_df), len(authors_df))

        # Check rows that should store same uuid, i.e., same author
        self.assertEqual(len(enriched_df[enriched_df['name'] == 'pepe']), 2)
        self.assertEqual(len(enriched_df[enriched_df['name'] == 'John Rambo']), 3)

        pepe_df = enriched_df[enriched_df['name'] == 'pepe']
        pepe_uuid = '0007b2fd4dbb5090a848c50717966a15a1772112'
        self.assertEqual(pepe_df.iloc[[0]]['uuid'].item(), pepe_uuid)
        self.assertEqual(pepe_df.iloc[[1]]['uuid'].item(), pepe_uuid)

        john_df = enriched_df[enriched_df['name'] == 'John Rambo']
        john_uuid = '000eabca391efd736dd8438e4d220b3f00808065'
        self.assertEqual(john_df.iloc[[0]]['uuid'].item(), john_uuid)
        self.assertEqual(john_df.iloc[[1]]['uuid'].item(), john_uuid)
        self.assertEqual(john_df.iloc[[2]]['uuid'].item(), john_uuid)

    def test_Onion(self):
        """Test several cases for the Onion analysis
        """

        members_df = pandas.read_csv(os.path.join(self.__enrich_dir,
                                                  "onion.csv"))
        onion = Onion(members_df)
        enriched_df = onion.enrich("author", "events")

        self.assertFalse(enriched_df.empty)

        self.assertTrue(len(enriched_df), 7)
        self.assertTrue(len(enriched_df[enriched_df["onion_role"] == "core"]), 1)
        self.assertTrue(len(enriched_df[enriched_df["onion_role"] == "regular"]), 3)
        self.assertTrue(len(enriched_df[enriched_df["onion_role"] == "casual"]), 4)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2019 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Daniel Izquierdo Cortazar <dizquierdo@bitergia.com>
#   Alberto Perez <alpgarcia@bitergia.com>
#


import pandas

import numpy as np

import re


class Enrich(object):
    """ Class that enriches information for a given dataset.

    Its usual way of working consists of providing a new columns
    in the dataframe from a given entry. This allows to keep
    adding extra information to the basics coming from the original
    data source.

    An example of this would be the addition of the affiliation,
    gender, workload adequacy and other metrics for a given commmit.
    """


class PairProgramming(Enrich):
    """ This class splits a commit into 2 of them according to the author name.

    There are cases where two developers have participated in a commit.
    One of the ways to identify those is claiming that commit as author, but
    also as a committer.

    Depending on the process of the organization, this may indicate a review
    process, a peer development process or even nothing.

    This class should be called when committer an author usually participate
    in the development of such commit.
    """

    def __init__(self, commits):
        """ Main constructor of the class

        :param commits: original list of commits
        :type commits: dataframe with commit information
        """

        self.commits = commits

    def enrich(self, column1, column2):
        """ This class splits those commits where column1 and column2
        values are different

        :param column1: column to compare to column2
        :param column2: column to compare to column1
        :type column1: string
        :type column2: string

        :returns: self.commits with duplicated rows where the values at
            columns are different. The original row remains while the second
            row contains in column1 and 2 the value of column2.
        :rtype: pandas.DataFrame
        """

        if column1 not in self.commits.columns or \
           column2 not in self.commits.columns:
            return self.commits

        # Select rows where values in column1 are different from values in column2
        pair_df = self.commits[self.commits[column1] != self.commits[column2]].copy()

        # Update values from column2
        pair_df[column1] = pair_df[column2].tolist()

        # Concatenate the original dataframe with the new one
        return pandas.concat([self.commits, pair_df], ignore_index=True)


class FileType(Enrich):
    """ This class creates a new column with the file type
    """

    def __init__(self, data):
        """ Main constructor of the class where the original dataframe
        is provided

        :param data: original dataframe
        : type data: pandas.DataFrame
        """

        self.data = data

    def enrich(self, column):
        """ This method adds a new column depending on the extension
        of the file.

        :param column: column where the file path is found
        :type column: string

        :return: returns the original dataframe with a new column named as
                 'filetype' that contains information about its extension
        :rtype: pandas.DataFrame
        """

        if column not in self.data:
            return self.data

        # Insert a new column with default values
        self.data["filetype"] = 'Other'

        # Insert 'Code' only in those rows that are
        # detected as being source code thanks to its extension
        reg = "\.bazel$|\.bazelrc$|\.bzl$|\.c$|\.cc$|\.cp$|\.cpp$|\.cxx$|\.c\+\+$|" +\
              "\.go$|\.h$|\.js$|\.mjs$|\.java$|\.py$|\.rs$|\.sh$|\.tf$|\.ts$"
        self.data.loc[self.data[column].str.contains(reg), 'filetype'] = 'Code'

        return self.data


class FilePath(Enrich):
    """ This class creates new columns with:
            * File extension
            * File path (excluding file name)
            * File name (excluding directories)
    """

    def __init__(self, data):
        """ Main constructor of the class where the original dataframe
        is provided

        :param data: original dataframe
        : type data: pandas.DataFrame
        """

        self.data = data

    def enrich(self, column):
        """ This method splits file path in new columns allowing further
        filtering on those particular path parts:
            * File extension
            * File directory name (full path excluding file name)
            * File name (excluding directories)
            * Path list including each directory/file as a separated element

        :param column: column where the file path is found
        :type column: string

        :return: returns the original dataframe with new columns named
                 'file_ext', 'file_dir_name', 'file_name', 'path_list'
        :rtype: pandas.DataFrame
        """

        if column not in self.data:
            return self.data

        # Insert new columns
        self.data['file_name'] = \
            self.data.apply(lambda row: row[column][row[column].rfind('/') + 1:],
                            axis=1)
        self.data['file_ext'] = \
            self.data.apply(lambda row:
                            '' if (row.file_name.rfind('.') == -1)
                            else row.file_name[row.file_name.rfind('.') + 1:],
                            axis=1)
        # To get correct dir name:
        # *  Replace multiple consecutive slashes by just one
        self.data['file_dir_name'] = self.data[column].str.replace('/+', '/', regex=True)
        self.data['file_dir_name'] = \
            self.data.apply(lambda row:
                            row.file_dir_name if row.file_dir_name.startswith('/')
                            else '/' + row.file_dir_name,
                            axis=1)
        self.data['file_dir_name'] = \
            self.data.apply(lambda row:
                            row.file_dir_name[:row.file_dir_name.rfind('/') + 1],
                            axis=1)

        # Clean filepath for splitting path parts:
        # * Replace multiple consecutive slashes by just one
        # * Remove leading slash if any, to avoid str.split to add an empty
        #   string to the resulting list of slices
        # * Remove trailing slash if any, to avoid str.split to add an empty
        #   string to the resulting list of slices
        self.data['file_path_list'] = self.data[column].str.replace('/+', '/', regex=True)
        self.data['file_path_list'] = self.data.file_path_list.str.replace('^/', '', regex=True)
        self.data['file_path_list'] = self.data.file_path_list.str.replace('/$', '', regex=True)
        self.data['file_path_list'] = self.data.file_path_list.str.split('/')

        return self.data


class Projects(Enrich):
    """ This class adds project info based on a pre-processed dataset
    """

    def __init__(self, data):
        """ Main constructor of the class where the original dataframe
        is provided.

        :param data: original dataframe
        :type data: pandas.DataFrame
        """

        self.data = data

    def enrich(self, column, projects):
        """ This method adds a new column named as 'project'
        that contains information about the associated project
        that the event in 'column' belongs to.

        :param column: column with information related to the project
        :type column: string
        :param projects: information about item - project
        :type projects: pandas.DataFrame

        :returns: original data frame with a new column named 'project'
        :rtype: pandas.DataFrame
        """

        if column not in self.data.columns:
            return self.data

        self.data = pandas.merge(self.data, projects, how='left', on=column)

        return self.data


class MessageLogFlag(Enrich):
    """ This class adds specific events for the
    given message log body
    """

    FLAGS_REGEX = {'Patch by Blink': r'\s*Patch by (?P<value>.+)$',
                   'Patch by WebKit': r'\s*Patch by (?P<value>.+) on .+$',
                   'Reviewed by WebKit': r'\s*Reviewed by (?P<value>.+) on .+$'}

    def __parse_flags(self, body):
        """Parse flags from a message"""
        flags = []
        values = []
        lines = body.split('\n')
        for line in lines:
            for name in self.FLAGS_REGEX:
                m = re.match(self.FLAGS_REGEX[name], line)

                if m:
                    flags.append(name)
                    values.append(m.group("value").strip())

        # TODO: this should be more consistent. Either
        # returning a list of strings or strings
        if flags == []:
            flags = ""
            values = ""

        if len(flags) == 1:
            flags = flags[0]
            values = values[0]

        return flags, values

    def __init__(self, data):

        """ Main constructor of the class where the original dataframe
        is provided.

        :param data: original dataframe
        :type data: pandas.DataFrame
        """

        self.data = data

    def enrich(self, column):
        """ This method helps to identify flags in the message log.
        As some communities may use the log message for the code
        authorship, specifig flags/tags are used to determine
        some actions by the authors or reviewers such who is
        the author or the reviewer is.

        The list of supported flags are found in the FLAGS_REGEX
        variable. In addition to this, a flag usually has a
        related developer where her name and email address are
        specified. This is also covered by this flag analysis.

        :param column: column where the text to analyze is found
        :type data: string
        """

        if column not in self.data.columns:
            return self.data

        flags_list = []
        values_list = []
        # Assuming the index of the dataframe is an integer
        for i in list(range(len(self.data))):
            flags, values = self.__parse_flags(self.data[column][i])
            flags_list.append(flags)
            values_list.append(values)

        self.data["flags"] = flags_list
        self.data["values"] = values_list

        return self.data


class EmailFlag(Enrich):
    """ This class adds specific events for the given
    email body
    """

    FLAGS_REGEX = {
        'Acked-by': '^Acked-by:(?P<value>.+)$',
        'Cc': '^Cc:(?P<value>.+)',
        'Fixes': '^Fixes:(?P<value>.+)$',
        'From': '^[Ff]rom:(?P<value>.+)$',
        'Reported-by': '^Reported-by:(?P<value>.+)$',
        'Tested-by': '^Tested-by:(?P<value>.+)$',
        'Reviewed-by': '^Reviewed-by:(?P<value>.+)$',
        'Release-Acked-by': '^Release-Acked-by:(?P<value>.+)$',
        'Signed-off-by': '^Signed-off-by:(?P<value>.+)$',
        'Suggested-by': '^Suggested-by:(?P<value>.+)$'}

    def __parse_flags(self, body):
        """Parse flags from a message"""
        flags = []
        values = []
        lines = body.split('\n')
        for line in lines:
            for name in self.FLAGS_REGEX:
                m = re.match(self.FLAGS_REGEX[name], line)

                if m:
                    flags.append(name)
                    values.append(m.group("value").strip())

        if flags == []:
            flags = ""
            values = ""

        return flags, values

    def __init__(self, data):

        """ Main constructor of the class where the original dataframe
        is provided.

        :param data: original dataframe
        :type data: pandas.DataFrame
        """

        self.data = data

    def enrich(self, column):
        """ This method helps to identify flags in the emails.
        As some communities may use the mailing list for the code
        review process, specifig flags/tags are used to determine
        some actions by the reviewers such as the moment when a
        patch is approved to merged.

        The list of supported flags are found in the FLAGS_REGEX
        variable where. In addition to this, a flag usually has a
        related developer where her name and email address are
        specified. This is also covered by this flag analysis.

        :param column: column where the text to analyze is found
        :type data: string
        """

        if column not in self.data.columns:
            return self.data

        flags_list = []
        values_list = []
        # Assuming the index of the dataframe is an integer
        for i in list(range(len(self.data))):
            flags, values = self.__parse_flags(self.data[column][i])
            flags_list.append(flags)
            values_list.append(values)

        self.data["flags"] = flags_list
        self.data["values"] = values_list

        return self.data


class SplitEmailDomain(Enrich):
    """ This class returns a new column with the domain of the email
    analyzed
    """

    def __parse_email(self, email):
        """ This function returns the domain of a given email
        """

        try:
            return email.split('@')[1]
        except Exception:
            return "unknown"

    def __init__(self, data):

        """ Main constructor of the class where the original dataframe
        is provided.

        :param data: original dataframe
        :type data: pandas.DataFrame
        """

        self.data = data

    def enrich(self, column):
        """ This enricher returns the same dataframe
        with a new column named 'domain'.
        That column is the result of splitting the
        email address of another column. If there is
        not a proper email address an 'unknown'
        domain is returned.

        :param column: column where the text to analyze is found
        :type data: string
        """

        if column not in self.data.columns:
            return self.data

        self.data['domain'] = self.data[column].apply(lambda x: self.__parse_email(x))
        return self.data


class ToUTF8(Enrich):
    """ This class helps to migrate (or ignore) the strings to utf-8
    """

    def __remove_surrogates(self, s, method='replace'):
        """ Remove surrogates in the specified string
        """

        if isinstance(s, list) and len(s) == 1:
            if self.__is_surrogate_escaped(s[0]):
                return s[0].encode('utf-8', method).decode('utf-8')
            else:
                return ""
        if isinstance(s, list):
            return ""
        if not isinstance(s, str):
            return ""
        if self.__is_surrogate_escaped(s):
            return s.encode('utf-8', method).decode('utf-8')
        return s

    def __is_surrogate_escaped(self, text):
        """ Checks if surrogate is escaped
        """

        try:
            text.encode('utf-8')
        except UnicodeEncodeError as e:
            if e.reason == 'surrogates not allowed':
                return True
        return False

    def __init__(self, data):
        """ Main constructor

        :parama data: Original dataset
        :type data: pandas.DataFrame
        """

        self.data = data

    def enrich(self, columns):
        """ This method convert to utf-8 the provided columns

        :param columns: list of columns to convert to
        :type columns: list of strings
        :return: original dataframe with converted strings
        :rtype: pandas.DataFrame
        """

        for column in columns:
            if column not in self.data.columns:
                return self.data

        for column in columns:
            a = self.data[column].apply(self.__remove_surrogates)
            self.data[column] = a

        return self.data


class SplitEmail(Enrich):
    """ This class split the tuple 'name <email>' into 'name' and 'email'.
    This adds two new columns named as 'user' and 'email to the provided
    one
    """

    def __parse_addr(self, addr):
        """ Parse email addresses
        """

        from email.utils import parseaddr

        value = parseaddr(addr)
        return value[0], value[1]

    def __init__(self, data):
        """ Main constructor oft he class

        :param data: original dataframe
        :type data: pandas.DataFrame
        """

        self.data = data

    def enrich(self, column):
        """ This method creates two new columns: user and email.
        Those contain the information coming from the usual tuple of
        user <email> found in several places like in the mailing lists
        or the git repositories commit.

        :param column: column to be used for this parser
        :type column: string

        :returns: dataframe with two new columns
        :rtype: pandas.DataFrame
        """

        if column not in self.data.columns:
            return self.data

        self.data["user"], self.data["email"] = zip(*self.data[column].map(self.__parse_addr))

        return self.data


class SplitLists(Enrich):
    """ This class looks for lists in the given columns and append at the
    end of the dataframe a row for each entry in those lists.
    """

    def __init__(self, data):
        """ Main constructor of the class where the original dataframe
        is provided

        :param data: original dataframe
        :type data: pandas.DataFrame
        """

        self.data = data

    def enrich(self, columns):
        """ This method appends at the end of the dataframe as many
        rows as items are found in the list of elemnents in the
        provided columns.

        This assumes that the length of the lists for the several
        specified columns is the same. As an example, for the row A
        {"C1":"V1", "C2":field1, "C3":field2, "C4":field3}
        we have three cells with a list of four elements each of them:
        * field1: [1,2,3,4]
        * field2: ["a", "b", "c", "d"]
        * field3: [1.1, 2.2, 3.3, 4.4]

        This method converts each of the elements of each cell in a new
        row keeping the columns name:

        {"C1":"V1", "C2":1, "C3":"a", "C4":1.1}
        {"C1":"V1", "C2":2, "C3":"b", "C4":2.2}
        {"C1":"V1", "C2":3, "C3":"c", "C4":3.3}
        {"C1":"V1", "C2":4, "C3":"d", "C4":4.4}

        :param columns: list of strings
        :rtype pandas.DataFrame
        """

        for column in columns:
            if column not in self.data.columns:
                return self.data

        # Looking for the rows with columns with lists of more
        # than one element
        first_column = list(self.data[columns[0]])
        count = 0
        append_df = pandas.DataFrame()
        for cell in first_column:
            if len(cell) >= 1:
                # Interested in those lists with more
                # than one element
                df = pandas.DataFrame()
                # Create a dataframe of N rows from the list
                for column in columns:
                    df[column] = self.data.loc[count, column]
                # Repeat the original rows N times
                extra_df = pandas.DataFrame([self.data.loc[count]] * len(df))
                for column in columns:
                    extra_df[column] = list(df[column])
                append_df = append_df.append(extra_df, ignore_index=True)
                extra_df = pandas.DataFrame()

            count = count + 1

        self.data = self.data.append(append_df, ignore_index=True)

        return self.data


class MaxMin(Enrich):
    """ This class creates two new columns with the maximum and
    minimum value of the given column
    """

    def __init__(self, data):
        """ Main constructor of the class where the original dataframe
        is provided.

        :param data: original dataframe
        :type data: pandas.DataFrame
        """

        self.data = data

    def enrich(self, columns, groupby):
        """ This method calculates the maximum and minimum value
        of a given set of columns depending on another column.
        This is the usual group by clause in SQL.

        :param columns: list of columns to apply the max and min values
        :param groupby: column use to calculate the max/min values
        :type columns: list of strings
        """

        for column in columns:
            if column not in self.data.columns:
                return self.data

        for column in columns:
            df_grouped = self.data.groupby([groupby]).agg({column: 'max'})
            df_grouped = df_grouped.reset_index()
            df_grouped.rename(columns={column: 'max_' + column}, inplace=True)
            self.data = pandas.merge(self.data, df_grouped, how='left', on=[groupby])

            df_grouped = self.data.groupby([groupby]).agg({column: 'min'})
            df_grouped = df_grouped.reset_index()
            df_grouped.rename(columns={column: 'min_' + column}, inplace=True)
            self.data = pandas.merge(self.data, df_grouped, how='left', on=[groupby])

        return self.data


class Gender(Enrich):
    """ This class creates three new columns with the gender of
    the name provided
    """

    def __init__(self, data, key=None, gender_file=None):
        """ Main constructor of the class where the original dataframe
        is provided.

        :param data: original dataframe
        :param key: genderize key (optional)
        :param gender_file: file with gender info, used as cache
        :type data: pandas.DataFrame
        :type key: string
        :type gender_file: string (as filepath)
        """

        from genderize import Genderize

        self.data = data
        self.gender = {}  # init the name-gender dictionary
        self.key = key
        self.gender_file = gender_file

        # Init the genderize connection
        self.connection = Genderize()
        if self.key:
            self.connection = Genderize(api_key=self.key)

        if self.gender_file:
            # This file is used as cache for the gender info
            # This helps to avoid calling once and again to the API
            fd = open(gender_file, "r")
            lines = fd.readlines()
            fd.close()
            # TODO: fix hardcoded code when reading columns and using
            #      separators
            for line in lines:
                gender_data = line.split("\t")
                self.gender[gender_data[1]] = {"gender_analyzed_name": gender_data[1],
                                               "gender": gender_data[2]}

    def enrich(self, column):
        """ This method calculates thanks to the genderize.io API the gender
        of a given name.

        This method initially assumes that for the given
        string, only the first word is the one containing the name
        eg: Daniel Izquierdo <dizquierdo@bitergia.com>, Daniel would be the name.

        If the same class instance is used in later gender searches, this stores
        in memory a list of names and associated gender and probability. This is
        intended to have faster identifications of the gender and less number of
        API accesses.

        :param column: column where the name is found
        :type column: string

        :return: original dataframe with four new columns:
         * gender: male, female or unknown
         * gender_probability: value between 0 and 1
         * gender_count: number of names found in the Genderized DB
         * gender_analyzed_name: name that was sent to the API for analysis
        :rtype: pandas.DataFrame
        """

        if column not in self.data.columns:
            return self.data

        splits = self.data[column].str.split(" ")
        splits = splits.str[0]
        self.data["gender_analyzed_name"] = splits.fillna("noname")
        self.data["gender_probability"] = 0
        self.data["gender"] = "Unknown"
        self.data["gender_count"] = 0

        names = list(self.data["gender_analyzed_name"].unique())

        for name in names:
            if name in self.gender.keys():
                gender_result = self.gender[name]
            else:
                try:
                    # TODO: some errors found due to encode utf-8 issues.
                    # Adding a try-except in the meantime.
                    gender_result = self.connection.get([name])[0]
                except Exception:
                    continue

                # Store info in the list of users
                self.gender[name] = gender_result

            # Update current dataset
            if gender_result["gender"] is None:
                gender_result["gender"] = "NotKnown"
            self.data.loc[self.data["gender_analyzed_name"] == name, 'gender'] =\
                gender_result["gender"]
            if "probability" in gender_result.keys():
                self.data.loc[self.data["gender_analyzed_name"] == name,
                              'gender_probability'] = gender_result["probability"]
                self.data.loc[self.data["gender_analyzed_name"] == name,
                              'gender_count'] = gender_result["count"]

        self.data.fillna("noname")
        return self.data


class TimeDifference(Enrich):
    """ This class creates a new column with the difference in seconds
    between two dates.
    """

    def __init__(self, data):
        """ Main constructor of the class where the original dataframe
        is provided

        :param data: original dataframe
        :type data: pandas.DataFrame
        """

        self.data = data

    def enrich(self, column1, column2):
        """ This method calculates the difference in seconds between
            the 2 columns (column2 - column1)

        The final result may provided negative values depending on the values
        from column1 and column2.

        :param column1: first column. Values in column1 must be datetime type
        :param column2: second column. Values in column2 must be datetime type
        :type column1: string
        :type column2: string

        :return: original dataframe with a new column with the difference
            between column2 - column1
        :rtype: pandas.DataFrame
        """

        if column1 not in self.data.columns or \
           column2 not in self.data.columns:
            return self.data

        self.data["timedifference"] = (self.data[column2] - self.data[column1]) / np.timedelta64(1, 's')
        return self.data


class Uuid(Enrich):
    """ This class adds new columns with the uuid of a given identity. If more
    not common columns (those not used to decide how to merge rows) are
    provided together with the uuid within the CSV file, all of them will also
    be merged in the resulting dataframe.
    """

    def __init__(self, data, file_path='data/uuids.csv',
                 drop_columns=[], drop_duplicates=[]):
        """ Main constructor of the class where the original dataframe
        is provided and the dataframe containing identities and their
        uuids is loaded from CSV file.

        :param data: original dataframe
        :param file_path: uuids file path (optional)
        :param drop_columns: columns to remove from the csv
        :param drop_duplicates: columns to use to remove duplicates
        :type data: pandas.DataFrame
        :type file_path: string
        """

        self.data = data

        # Read csv to data frame, read '\N' (null in MySQL export format) also
        # as NaN (this is the way pandas deal with null values)
        self.uuids_df = pandas.read_csv(filepath_or_buffer=file_path, na_values='\\N', sep=',')
        # Remove required columns to later merge
        for column in drop_columns:
            self.uuids_df.drop(column, 1)

        if len(drop_duplicates) > 0:
            self.uuids_df.drop_duplicates(subset=drop_duplicates, inplace=True)
        else:
            self.uuids_df.drop_duplicates(inplace=True)

    def enrich(self, columns):
        """ Merges the original dataframe with corresponding entity uuids based
        on the given columns. Also merges other additional information
        associated to uuids provided in the uuids dataframe, if any.

        :param columns: columns to match for merging
        :type column: string array

        :return: original dataframe with at least one new column:
         * uuid: identity unique identifier
        :rtype: pandas.DataFrame
        """

        for column in columns:
            if column not in self.data.columns:
                return self.data

        self.data = pandas.merge(self.data, self.uuids_df, how='left', on=columns)
        self.data = self.data.fillna("notavailable")

        return self.data


class Onion(Enrich):
    """ This class adds a new column with the role of the author with respect
    to the amount of work in a given column. The onion model is based on the
    analysis of the community where the 80% of the work is done by member.
    """

    def __init__(self, data):
        """ Main constructor of the class where the original dataframe
        is provided and the dataframe containing identities and their
        uuids is loaded from CSV file.

        :param data: original dataframe
        :param file_path: uuids file path (optional)
        :param drop_columns: columns to remove from the csv
        :param drop_duplicates: columns to use to remove duplicates
        :type data: pandas.DataFrame
        :type file_path: string
        """

        self.data = data

    def enrich(self, member_column, events_column):
        """ Calculates the onion model for the given set of columns.
        This expects two columns as input (typically the author and
        the amount of events) and returns a third column with the
        role (core, regular, casual) of such community member.

        :param columns: columns to match for calculating the onion model
        :type column: string array

        :return: original dataframe with three new columns and ordered by role
                 importance:
         * onion_role: "core", "regular", or "casual"
         * percent_cum_net_sum: percentage of the activity up to such developer
         * cum_net_sum: accumulated activity up to such developer
        :rtype: pandas.DataFrame
        """

        if member_column not in self.data.columns or \
           events_column not in self.data.columns:
            return self.data

        # Order the data... just in case
        self.data.sort_values(by=events_column, ascending=False, inplace=True)
        # Reset the index to properly work with other methods
        self.data.reset_index(inplace=True)
        # Remove resultant new 'index' column
        self.data.drop(["index"], axis=1, inplace=True)

        # Calculate onion limits and accumulative sum and percentage
        self.data["cum_net_sum"] = self.data[events_column].cumsum()
        self.data["percent_cum_net_sum"] = (self.data.cum_net_sum / self.data[events_column].sum()) * 100

        # Assign roles based on the percentage
        self.data["onion_role"] = pandas.cut(self.data["percent_cum_net_sum"],
                                             [0.0, 80.0, 95.0, 100.0],
                                             labels=["core", "regular", "casual"])

        return self.data
